/* $Id: freqspec.hpp 61 2006-07-31 21:15:25Z pjohnson $
  ------------------------------------------------------------------------
  This file is part of the PIIM package; please see:
    Johnson, PLF and Slatkin, M. 2006. "Inference of population
    genetic parameters in metagenomics: a clean look at messy data."
    Genome Research (in press).
 
  Copyright 2006 The Regents of the University of California.  All
  rights reserved.
 
  IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
  INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
  DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGE.
 
  REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING
  DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
  IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 
  This software is distributed under the terms of Version 2 of the
  GNU General Public License as published by the Free Software
  Foundation.  If you have not received a copy of Version 2 of the
  GNU General Public License with this program, you may download the
  GPL from the following url: http://www.gnu.org/copyleft/gpl.html.

  ---------------------------------------------------------------------------
  Classes to generate various frequency spectra
  ---------------------------------------------------------------------------
*/

#include "util.hpp"
#include "gsl_util.hpp"
#include <vector>
#include <iostream>

#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf_gamma.h>
#include <time.h> //for initializing rng

using namespace std;

//---------------------------------------------------------------------------
// base class for frequency spectrum
class CFreqSpectrum {
public:
    typedef vector< pair<double, bool> > TParams;

    CFreqSpectrum(void) { m_Folded = true;}
    virtual ~CFreqSpectrum(void) {};
    virtual const string GetName(void) const = 0;
    virtual CFreqSpectrum* Clone(void) const = 0;
    virtual bool ParamsOutsideBounds(void) const { return false; }

    double& SetParam(unsigned int n) {
        if (m_Params.size() <= n) { m_Params.resize(n+1); }
        m_Cache.Reset();
        m_Params[n].second = true; return m_Params[n].first; }
    void    ClearParam(unsigned int n) {
        m_Cache.Reset(); m_Params[n].second = false; }

    double  GetParam(unsigned int n) const {
        assert(n < m_Params.size()); return m_Params[n].first; }
    bool    HasParam(unsigned int n) const {
        return n < m_Params.size()  &&  m_Params[n].second; }
    unsigned int GetNumParams(void) const {
        int c = 0;
        for (TParams::const_iterator it = m_Params.begin();
             it != m_Params.end();  ++it) {
            if (it->second) { ++c; }
        }
        return c;
    }

    void SetParams(const string &paramStr);
    TParams& SetParams(void) { m_Cache.Reset(); return m_Params; }
    const TParams& GetParams(void) const { return m_Params; }

    bool IsFolded(void) const { return m_Folded; }
    void SetFolded(bool f) { m_Folded = f; m_Cache.Reset(); }

    double operator() (int numMinor, int depth) const {
        try { //cache HERE to avoid a vtable lookup (or two, if folded)
            return m_Cache[depth][numMinor];
        } catch (exception) {} //missed cache

        if (numMinor == depth) {
            //symmetry simplifies calculations
            return m_Cache.Set(depth).Set(numMinor) =
                m_Folded ? x_Calc(0, depth)/2 : 0;
        }
        assert(numMinor < depth);
        if (m_Folded  &&  /*numMinor != depth - numMinor  &&*/  numMinor > 0) {
            return m_Cache.Set(depth).Set(numMinor) =
                (x_Calc(numMinor, depth) + x_Calc(depth-numMinor, depth))/2.0;
        } else {
            return m_Cache.Set(depth).Set(numMinor) = m_Folded ? 
                x_Calc(numMinor, depth) / 2 :
                x_Calc(numMinor, depth);
        }
    }

    void PrintSpectrum(int depth) const { //primarily for debugging
        for (int i = 0;  i < depth;  ++i) {
            cerr << (*this)(i, depth) << " ";
        }
    }

    friend ostream& operator<< (ostream &o, const TParams &p) {
        for (TParams::const_iterator it = p.begin();  it != p.end();  ++it) {
            if (it->second) {
                if (it != p.begin()) {
                    o << "\t";
                }
                o << it->first;
            }
        }
        return o;
    }

protected:
    virtual double x_Calc (int numMinor, int depth) const = 0;

    bool m_Folded;
    TParams m_Params;

private:
    mutable CCache< CCache< double > > m_Cache;
};

//---------------------------------------------------------------------------

class CFSNeutral : public CFreqSpectrum {
public:
    enum EParams {
        eTheta = 0,
        eFixed = 1
    };
    CFSNeutral(void) {
        SetParam(eTheta) = .01;
    }
    virtual const string GetName(void) const { return "Neutral"; }
    void SetUseFixed(bool f) {
        if (f) {
            SetParam(eFixed) = .001;
        } else {
            ClearParam(eFixed);
        }
    }
    virtual bool ParamsOutsideBounds(void) const {
        double norm = 0;
        for (int i = 1;  i <= 30;  ++i) {//FIXME: random max depth
            norm += 1.0/i;
        }
        return (GetParam(eTheta) < 0  ||  GetParam(eTheta) * norm > 1  ||
                (HasParam(eFixed)  &&
                 (GetParam(eFixed) < 0  ||
                  GetParam(eFixed) > GetParam(eTheta))));
    }
    virtual CFreqSpectrum* Clone(void) const { return new CFSNeutral(*this); }
protected:
    virtual double x_Calc (int numMinor, int depth) const;
};

//---------------------------------------------------------------------------

class CFSGrowth : public CFreqSpectrum {
public:
    enum EParams {
        eTheta = 0,
        eGrowthRate = 1
    };
    CFSGrowth(void) {
        SetParam(eTheta) = 0.01;
        SetParam(eGrowthRate) = 0.05;
    }
    virtual const string GetName(void) const { return "Growth"; }
    virtual bool ParamsOutsideBounds(void) const {
        return (GetParam(eTheta) <= 0  ||  GetParam(eGrowthRate) < 1e-10);
    }
    virtual CFreqSpectrum* Clone(void) const {return new CFSGrowth(*this);}
protected:
    virtual double x_Calc (int numMinor, int depth) const;

private:
    double x_W(int b, int j, int n) const;
    double x_V(int j, int n) const;
    double x_Fac(int start, int stop) const;
    double x_E(int j) const;

    mutable CCache< CCache< CCache<double> > > m_WCache;
};
