/* $Id: simulate_data.cpp 65 2006-08-24 23:55:08Z pjohnson $
  ------------------------------------------------------------------------
  This file is part of the PIIM package; please see:
    Johnson, PLF and Slatkin, M. 2006. "Inference of population
    genetic parameters in metagenomics: a clean look at messy data."
    Genome Research (in press).
 
  Copyright 2006 The Regents of the University of California.  All
  rights reserved.
 
  IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT,
  INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING
  LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
  DOCUMENTATION, EVEN IF REGENTS HAS BEEN ADVISED OF THE POSSIBILITY
  OF SUCH DAMAGE.
 
  REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
  FOR A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING
  DOCUMENTATION, IF ANY, PROVIDED HEREUNDER IS PROVIDED "AS
  IS". REGENTS HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
  UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 
  This software is distributed under the terms of Version 2 of the
  GNU General Public License as published by the Free Software
  Foundation.  If you have not received a copy of Version 2 of the
  GNU General Public License with this program, you may download the
  GPL from the following url: http://www.gnu.org/copyleft/gpl.html.

  ---------------------------------------------------------------------------
  Simulates metagenomic-style data complete with sequencing errors
  ---------------------------------------------------------------------------
  Inputs: empirical q-score / depth distributions.

*/

#include "util.hpp"
#include "freqspec.hpp"
#include <libxml++/libxml++.h>
#include <fstream>
using namespace std;

enum EBaseType {
    eMajor = 0,
    eMinor = 1,
    eOther = 2
};

typedef vector<double> TProbs;
typedef vector< pair<unsigned int,double> > TProbDist;

class CExceptionTriallelic : exception {
public:
    CExceptionTriallelic(bool fixed) : m_Fixed(fixed) {}
    bool m_Fixed;
};

class CDataGenerator {
public:
    CDataGenerator(bool biAllelic, bool exact) {
        m_RNG = gsl_rng_alloc(gsl_rng_mt19937);
        gsl_rng_set(m_RNG, time(NULL));
        m_BiAllelic = biAllelic;//whether to act like there are only 2 nucletides (otherwise correctly model all 4)
        m_Exact = exact; //whether to attempt to use exact FS (aside from rounding errors) or pull stochastically
    }
    ~CDataGenerator(void) { gsl_rng_free(m_RNG); }
    static void InitializeDists(istream &scores, istream &depths);

    void Generate(unsigned int totalSites, const CFreqSpectrum &fs,
                  xmlpp::Element *root);

protected:
    void x_GetErrorProbs(unsigned int depth, TProbs &probs);
    void x_CreateSite(unsigned int numMinor, unsigned int depth);
    int x_DrawValue(const TProbDist &probs) const;
    double x_CalcAverageDepth(void) const;

private:
    gsl_rng *m_RNG;
    xmlpp::Element *m_XML;
    unsigned int m_ContigPos;
    bool m_BiAllelic, m_Exact;
    static TProbDist sm_ErrDist, sm_DepthDist;
};

TProbDist CDataGenerator::sm_ErrDist, CDataGenerator::sm_DepthDist;

/*---------------------------------------------------------------------------*/
// **DUMMY** class for m.s.
class CFSms : public CFreqSpectrum {
public:
    typedef map<unsigned int, TProbDist> TProbDists;
    enum EParams {
        eTheta = 0,
        eGrowth = 1,
        eRho = 2
    };

    void GetSpectrum(int numSites, double aveDepth, TProbDists& pd) const {
        ostringstream oss;
        oss << "./ms2spectrum.pl "
            << "-d " << aveDepth << " "
            << "-n " << numSites << " "
            << "-t " << GetParam(eTheta) << " "
            << "-r " << GetParam(eRho) << " "
            << "-g " << GetParam(eGrowth);

        FILE *MS;
        MS = popen(oss.str().c_str(), "r");

        int depth;
        TProbDist depthDist;

        while (fscanf(MS, "%d", &depth) == 1) {
            for (int i = 0;  i < depth;  ++i) {
                double p;
                assert(fscanf(MS, "%lf", &p) == 1);
                pd[depth].push_back(TProbDist::value_type(i, p));
            }
        }
        pclose(MS);
    }

    virtual const string GetName(void) const { return "m.s."; }
    virtual bool ParamsOutsideBounds(void) const { return false; }
    virtual CFreqSpectrum* Clone(void) const { return new CFSms(*this); }
protected:
    virtual double x_Calc (int, int) const {return 0;};
};

/*---------------------------------------------------------------------------*/
// PRE : probability of sequencing error
// POST: quality "score"
inline unsigned int Err2Score(double errorProb) {
    return (unsigned int) rint(log10(errorProb) * -10);
}
/*---------------------------------------------------------------------------*/
// PRE : quality "score"
// POST: probability of sequencing error
inline double Score2Err(unsigned int score) {
    return pow(10, -((double)score)/10.0);
}

/*---------------------------------------------------------------------------*/
// PRE : input streams containg histograms of scores and depths
// POST: static members initialized
void CDataGenerator::InitializeDists(istream &scoreIn, istream &depthIn) {
    unsigned int i;
    double v;

    scoreIn >> i >> v;
    while (scoreIn.good()) {
        sm_ErrDist.push_back(TProbDist::value_type(i,v));
        scoreIn >> i >> v;
    }

    depthIn >> i >> v;
    while (depthIn.good()) {
        sm_DepthDist.push_back(TProbDist::value_type(i,v));
        depthIn >> i >> v;
    }
}

/*---------------------------------------------------------------------------*/
// PRE : discrete probability distribution
// POST: a random value according to the prob dist
int CDataGenerator::x_DrawValue(const TProbDist &probs) const {
    double r = gsl_rng_uniform(m_RNG);
    
    TProbDist::const_iterator it = probs.begin();
    assert(it != probs.end());
    for (double cur = it->second;  it != probs.end()  &&  cur < r;
         ++it, cur += it->second);
    if (it == probs.end()) {
        --it;
    }
    
    return it->first;
}

/*---------------------------------------------------------------------------*/
// PRE : depth of sample (and thus # of error probs to generate)
// POST: (randomized) error probabilities for each chromosome in sample
void CDataGenerator::x_GetErrorProbs(unsigned int depth, TProbs &probs) {
    probs.resize(depth);
    for (unsigned int i = 0;  i < depth;  ++i) {
        probs[i] = Score2Err(x_DrawValue(sm_ErrDist));
    }
}

/*---------------------------------------------------------------------------*/
// PRE : number of true minor alleles & sample depth at site
// POST: counts for the observed # minor, major & other alleles
void CDataGenerator::x_CreateSite(unsigned int numMinor, unsigned int depth) {
    TProbs errorProbs;
    vector<bool> minor, error, other;
    minor.resize(depth);
    error.resize(depth);
    other.resize(depth);

    x_GetErrorProbs(depth, errorProbs);

    vector<unsigned char> site(depth);
    vector<unsigned int> alleles(4);
    for (unsigned int i = 0;  i < depth;  ++i) {
        double r = gsl_rng_uniform(m_RNG);
        double errorProb = errorProbs[i];

        if (i < numMinor) { //true minor allele
            if (r < errorProb) { //error!
                error[i] = true;
                if (m_BiAllelic) {
                    site[i] = 0;
                } else {
                    if (r < (errorProb / 3)) {
                        site[i] = 0;
                    } else if (r < (2*errorProb / 3)) {
                        site[i] = 2;
                    } else {
                        site[i] = 3;
                    }
                }
            } else {
                error[i] = false;
                site[i] = 1;
            }
        } else { //true major allele
            if (r < errorProb) {
                error[i] = true;
                if (m_BiAllelic) {
                    site[i] = 1;
                } else {
                    if (r < (errorProb / 3)) {
                        site[i] = 1;
                    } else if (r < (2*errorProb / 3)) {
                        site[i] = 2;
                    } else {
                        site[i] = 3;
                    }
                }
            } else {
                error[i] = false;
                site[i] = 0;
            }
        }
        ++alleles[site[i]];
    }


    enum EStates {
        eAncestral=0, eDerived, eOther1, eOther2
    };
    EStates mapping[4];// = {eAncestral, eDerived, eDerived, eDerived};
    // = {eAncestral, eDerived, eOther1, eOther2};
    /*
    int mx = 0;
    for (unsigned int i = 0;  i < 4;  ++i) {
        if (alleles[i] > alleles[mx]) {
            mx = i;
        }
    }
    for (unsigned int i = 0;  i < 4;  ++i) {
        mapping[i] = (i == mx) ? eAncestral:eDerived;
    }
    */

    int state = mapping[0] = eAncestral;
    for (unsigned int i = 1;  i < 4;  ++i) {
        if (alleles[i] > 0) {
            mapping[i] = (EStates) ++state;
        }
    }
    
    xmlpp::Element *column = m_XML->add_child("Column");
    column->set_attribute("pos", IntToString(++m_ContigPos));
    for (unsigned int chrI = 0;  chrI < depth;  ++chrI) {
        xmlpp::Element *chr = column->add_child("Chr");
        chr->set_attribute("score", IntToString(Err2Score(errorProbs[chrI])));
        switch(mapping[site[chrI]]) {
            case eAncestral : break;
            case eDerived : chr->set_attribute("minor", ""); break;
            case eOther1 : chr->set_attribute("other", "1"); break;
            case eOther2 : chr->set_attribute("other", "2"); break;
        }
        if (error[chrI]) {
            chr->set_attribute("error", "");
        }
    }

    if (state > eDerived) {
        throw CExceptionTriallelic(numMinor == 0);
    }
}

/*---------------------------------------------------------------------------*/
// PRE : sm_DepthDist initialized
// POST: average depth
double CDataGenerator::x_CalcAverageDepth(void) const {
    double ave = 0;

    for (TProbDist::const_iterator depthI = sm_DepthDist.begin();
         depthI != sm_DepthDist.end();  ++depthI) {
        ave += depthI->first * depthI->second;
    }

    return ave;
}

/*---------------------------------------------------------------------------*/
// PRE : number of sites to generate, frequency spectrum to use, root
// node to XML document
// POST: sites generated & dumped into XML
void CDataGenerator::Generate(unsigned int totalSites, const CFreqSpectrum &fs,
                              xmlpp::Element *root) {
    m_XML = root;
    m_ContigPos = 0;
    unsigned int triallelicFCnt = 0, triallelicPCnt = 0;

    //special case m.s. because depth is simulated as well
    if (fs.GetName() == "m.s.") {
        CFSms::TProbDists fullSpectrum;
        dynamic_cast<const CFSms&>(fs).GetSpectrum
            (totalSites, x_CalcAverageDepth(), fullSpectrum);
        for (CFSms::TProbDists::const_iterator depthI = fullSpectrum.begin();
             depthI != fullSpectrum.end();  ++depthI) {

#ifdef _DEBUG
                cerr << depthI->first << ": ";
#endif

            for (unsigned int freqI = 0;  freqI < depthI->first;  ++freqI) {
                unsigned int numAtFreq = (int) (totalSites * depthI->second[freqI].second);
#ifdef _DEBUG
                cerr << numAtFreq << ", ";
#endif
                //create numAtFreq sites with depthI, freqI
                for (unsigned int siteI = 0;  siteI < numAtFreq;  ++siteI) {
                    try {
                        x_CreateSite(freqI, depthI->first);
                    } catch (CExceptionTriallelic& e ) {
                        if (e.m_Fixed) {
                            ++triallelicFCnt;
                        } else {
                            ++triallelicPCnt;
                        }
                    }
                }
            }
#ifdef _DEBUG
            cerr << endl;
#endif
        }
        return;
    }

    //normal frequency spectrum (i.e. non-M.S.)
    for (TProbDist::const_iterator depthI = sm_DepthDist.begin();
         depthI != sm_DepthDist.end();  ++depthI) {        
        unsigned int numAtDepth = (unsigned int)rint(depthI->second * totalSites);
        if (m_Exact) { //use asymptotically exact spectrum
            for (unsigned int freqI = 0;  freqI < depthI->first;  ++freqI) {
                unsigned int numAtFreq = (int)
                    rint(fs(freqI, depthI->first) * numAtDepth);
#ifdef _DEBUG
                cerr << depthI->first << "/" << numAtDepth << " " << freqI << " -- " << numAtFreq << "\t" << fs(freqI, depthI->first) << "%" << endl;
#endif
                //create numAtFreq sites with depthI, freqI
                for (unsigned int siteI = 0;  siteI < numAtFreq;  ++siteI) {
                    try {
                        x_CreateSite(freqI, depthI->first);
                    } catch (CExceptionTriallelic& e ) {
                        if (e.m_Fixed) {
                            ++triallelicFCnt;
                        } else {
                            ++triallelicPCnt;
                        }
                    }
                }
            }

        } else { //draw stochastically from distribution
            TProbDist spectrum;
            for (unsigned int freqI = 0;  freqI < depthI->first;  ++freqI) {
                rint(fs(freqI, depthI->first) * numAtDepth);
                spectrum.push_back(TProbDist::value_type
                                   (freqI,fs(freqI, depthI->first)));
            }

            for (unsigned int siteI = 0;  siteI < numAtDepth;  ++siteI) {
                try {
                    x_CreateSite(x_DrawValue(spectrum), depthI->first);
                } catch (CExceptionTriallelic& e) {
                    if (e.m_Fixed) {
                        ++triallelicFCnt;
                    } else {
                        ++triallelicPCnt;
                    }
                }
            }
        }
    }

#ifdef _DEBUG
    cerr << "# of triallelic sites = " << triallelicFCnt + triallelicPCnt
         << " (" << triallelicFCnt << "," << triallelicPCnt << ")" << endl;
#endif
}

/*---------------------------------------------------------------------------*/
// PRE : 
// POST: 
int Main(TOpts &opts) {
    const unsigned int totalSites = atoi(opts['n']);
    {
        ifstream scoresIn(opts.find('s') != opts.end() ?
                          opts['s'] : "score_dist.dat");
        ifstream depthsIn(opts.find('d') != opts.end() ? 
                          opts['d'] : "depth_dist.dat");
        CDataGenerator::InitializeDists(scoresIn, depthsIn);
    }

    auto_ptr<CFreqSpectrum> fs;
    if (string(opts['f']) == "neutral") {
        fs.reset(new CFSNeutral);        
    } else if (string(opts['f']) == "growth") {
        fs.reset(new CFSGrowth);
    } else if (string(opts['f']) == "ms") {
        fs.reset(new CFSms);
    } else {
        throw runtime_error(string("Invalid frequency spectrum '") +
                            opts['f'] + "'.");
    }
    if (opts.find('p') != opts.end()) {
        fs->SetParams(opts['p']);
    }
    fs->SetFolded(false);

    xmlpp::Document doc;
    doc.set_internal_subset("Contig", "", "qual.dtd");
    xmlpp::Element *root = doc.create_root_node("Contig");

    CDataGenerator gen(opts.find('b') != opts.end(),
                       opts.find('e') != opts.end());
    gen.Generate(totalSites, *fs, root);

    xmlpp::Element *params = root->add_child("FSParams");
    params->set_attribute("dist", fs->GetName());
    string paramStr;
    for (unsigned int paramI = 0;  paramI < fs->GetNumParams();  ++paramI) {
        paramStr += DblToString(fs->GetParam(paramI)) + " ";
    }
    params->set_attribute("params", paramStr);
    cout << doc.write_to_string_formatted();

    return 0;
}

/*---------------------------------------------------------------------------*/
int main(int argc, char** argv) {
    try {
        TOpts opts;
        x_ProcessOptions
            (argc, argv, "f:ben:s:d:p:", "fpn", opts,
             " -f <neutral|growth> [-b] [-n <number of sites>] [-s <scorefile>] [-d <depthfile>] [-p '<param string>']\n"
             "Where\t-f --> frequency spectrum to use\n"
             "\t-p --> string containing parameters\n"
             "\t-n --> number of sites to generate\n"
             "\t-s <filename>--> (default: score_dist.dat) file from which to load score dist\n"
             "\t-d <filename>--> (default: depth_dist.dat) file from which to load depth dist\n"
             "\t-b --> (optional) flag to simulate biallelic\n"
             "\t-e --> (optional) flag to force exact spectrum (otherwise stochastic)\n");

        return Main(opts);
    } catch (exception &e) {
        cerr << "Exception:\n\t" << e.what() << endl;
        return 1;
    }
}

